const kUserDefined = "{searchTerms}";
const kIsUserInput = /(\s|["'])user(\s|[>="']|$)/i;

function LOG(aText) {
  //alert("*** Sherlock parser: " + aText + "\n");
  var consoleService = Cc["@mozilla.org/consoleservice;1"].
                       getService(Ci.nsIConsoleService);
  consoleService.logStringMessage(aText);
}

function Engine(aFile) {
  this._file = aFile;
  this._data = readFile(aFile);

  this._name = "";
  this._queryCharset = "";
  this._description = "";
  this._urls = [];

  try {
    this.parse();
  } catch (ex) {
    LOG("Error parsing engine:" + aFile.leafName + "\n" + ex);
    throw Cr.NS_ERROR_FAILURE;
  }
}
Engine.prototype = {
  parse: function () {
    /**
     * Trims leading and trailing whitespace from aStr.
     */
    function sTrim(aStr) {
      return aStr.replace(/^\s+/g, "").replace(/\s+$/g, "");
    }

    /**
     * Extracts one Sherlock "section" from aSource. A section is essentially
     * an HTML element with attributes, but each attribute must be on a new
     * line, by definition.
     *
     * @param aLines
     *        An array of lines from the sherlock file.
     * @param aSection
     *        The name of the section (e.g. "search" or "browser"). This value
     *        is not case sensitive.
     * @returns an object whose properties correspond to the section's
     *          attributes.
     */
    function getSection(aLines, aSection) {
      //LOG("Sherlock lines:\n" + aLines.join("\n"));
      aSection  = aSection.toLowerCase();
      var lines = aLines;
      var startMark = new RegExp("^\\s*<" + aSection + "\\s*", "gi");
      var endMark   = /\s*>\s*$/gi;

      var foundStart = false;
      var startLine, numberOfLines;
      // Find the beginning and end of the section
      for (var i=0; i<lines.length; i++) {
        if (foundStart) {
          if (endMark.test(lines[i])) {
            numberOfLines = i - startLine;
            // Remove the end marker
            lines[i] = lines[i].replace(endMark, "");
            // If the endmarker was not the only thing on the line, include
            // this line in the results
            if (lines[i])
              numberOfLines++;
            break;
          }
        } else {
          if (startMark.test(lines[i])) {
            foundStart = true;
            // Remove the start marker
            lines[i] = lines[i].replace(startMark, "");
            startLine = i;
            // If the line is empty, don't include it in the result
            if (!lines[i])
              startLine++;
          }
        }
      }
      //LOG("_parseAsSherlock::getSection:\nStart index: " + startLine +
      //    "\nNumber of lines: " + numberOfLines);
      lines = lines.splice(startLine, numberOfLines);
      LOG("_parseAsSherlock::getSection:\nSection lines:\n" +
          lines.join("\n"));

      var section = {};
      for (var i=0; i<lines.length; i++) {
        var line = sTrim(lines[i]);

        var els = line.split("=");

        if (!els[0] || !els[1])
          continue;

        // Strip leading and trailing whitespace, remove quotes from the
        // value, and remove any trailing ">" characters
        var name = sTrim(els[0].toLowerCase());
        var value = sTrim(els[1]).replace(/^["']/, "")
                                 .replace(/["']\s*>?\s*$/, "") || "";

        // Don't clobber existing attributes
        if (!(name in section))
          section[name] = value;
      }
      return section;
    }

    /**
     * Returns an array of name-value pair arrays representing the Sherlock
     * file's input elements. User defined inputs return kUserDefined as the
     * value. Elements are returned in the order they appear in the source
     * file.
     *
     *   Example:
     *      <input name="foo" value="bar">
     *      <input name="foopy" user>
     *   Returns:
     *      [["foo", "bar"], ["foopy", "{searchTerms}"]]
     */
    function getInputs(aLines) {

      /**
       * Extracts an attribute value from a given a line of text.
       *    Example: <input value="foo" name="bar">
       *      Extracts the string |foo| or |bar| given an input aAttr of
       *      |value| or |name|.
       * Attributes may be quoted or unquoted. If unquoted, any whitespace
       * indicates the end of the attribute value.
       *    Example: < value=22 33 name=44\334 >
       *      Returns |22| for "value" and |44\334| for "name".
       *
       * @param aAttr
       *        The name of the attribute for which to obtain the value. This
       *        value is not case sensitive.
       * @param aLine
       *        The line containing the attribute.
       *
       * @returns the attribute value, or an empty string if the attribute
       *          doesn't exist.
       */
      function getAttr(aAttr, aLine) {
        //LOG("Getting attr: \"" + aAttr + "\" for line: \"" + aLine + "\"");
        // We're not case sensitive, but we want to return the attribute value
        // in it's original case, so create a copy of the source
        var lLine = aLine.toLowerCase();
        var attr = aAttr.toLowerCase();

        var attrStart = lLine.indexOf(" " + attr);
        if (attrStart == -1) {
          // If this is the "user defined input" (i.e. contains the empty
          // "user" attribute), return our special keyword
          if (kIsUserInput.test(lLine) && attr == "value") {
            //LOG("_parseAsSherlock::getAttr: Found user input!\nLine:\"" + lLine
            //    + "\"");
            return kUserDefined;
          }
          // The attribute doesn't exist - ignore
          //LOG("_parseAsSherlock::getAttr: Failed to find attribute:\nLine:\""
          //    + lLine + "\"\nAttr:\"" + attr + "\"");
          return "";
        }

        var valueStart = lLine.indexOf("=", attrStart) + "=".length;
        if (valueStart == -1)
          return "";

        var quoteStart = lLine.indexOf("\"", valueStart);
        if (quoteStart == -1) {

          // Unquoted attribute, get the rest of the line, trimmed at the first
          // sign of whitespace. If the rest of the line is only whitespace,
          // returns a blank string.
          restOfLine = lLine.substr(valueStart).replace(/\s.*$/, "");
          return restOfLine;

        } else {
          // Adjust the start index to account for the opening quote
          valueStart = quoteStart + "\"".length;
          // Find the closing quote
          valueEnd = lLine.indexOf("\"", valueStart);
          // If there is no closing quote, just go to the end of the line
          if (valueEnd == -1)
            valueEnd = aLine.length;
        }
        return aLine.substring(valueStart, valueEnd);
      }

      var inputs = [];
      //LOG("_parseAsSherlock::getInputs:\nLines:\n" + lines);
      // Filter out everything but non-inputs
      lines = lines.filter(function (line) {
        return /^\s*<input/i.test(line);
      });
      //LOG("_parseAsSherlock::getInputs:\nFiltered lines:\n" + lines);
      lines.forEach(function (line) {
        // Strip leading/trailing whitespace and remove the surrounding markup
        // ("<input" and ">")
        line = sTrim(line).replace(/^<input/, "").replace(/>$/, "");

        // If this is one of the "directional" inputs (<inputnext>/<inputprev>)
        if (/^(prev|next)/.test(line)) {

          // Make it look like a normal input by removing "prev" or "next"
          line = line.replace(/^(prev|next)/, "");

          // If it has a name, give it a dummy value to match previous
          // nsInternetSearchService behavior
          if (/name\s*=/.test(line)) {
            line += " value=\"0\"";
          } else
            return; // Line has no name, skip it
        }

        var name = getAttr("name", line);
        var value = getAttr("value", line);
        //LOG("Got input:\nName:\"" + name + "\"\nValue:\"" + value + "\"");
        if (name && value)
          inputs.push([name, value]);
      });
      return inputs;
    }

    function err(aErr) {
      LOG("_parseAsSherlock:\nSherlock param error:\n" + aErr);
      throw Cr.NS_ERROR_FAILURE;
    }

    // Filter out all comments and whitespace-only lines
    var lines = this._data.split("\n").filter(function (line) {
      return (!(/^\s*$/i.test(line)) && !(/^\s*#/i.test(line)));
    });

    var searchSection = getSection(lines, "search");
    LOG("_parseAsSherlock:\nSearch section:\n" + searchSection.toSource());
    var inputs        = getInputs(lines);
    LOG("_parseAsSherlock:\nInputs:\n" + inputs.toSource());

    this._name = searchSection["name"] || err("Missing name!");
    this._description = searchSection["description"] || "";
    this._queryCharset = searchSection["querycharset"] ||
                            getCharSetFromCode(searchSection["queryencoding"]);

    // XXX should this really fall back to GET?
    var method = searchSection["method"] || "GET";
    var template = searchSection["action"] || err("Missing action!");

    var url = new EngineURL("text/html", method, template);
    for (var x in inputs)
      url.addParam(inputs[x][0], inputs[x][1]);
    this._urls.push(url);
  },
};

const kIllegalWords = /(\{count\})|(\{startIndex\})|(\{startPage\})|(\{language\})|(\{outputEncoding\})|(\{inputEncoding\})/;
const kValidWords = /\{searchTerms\}/gi;

function EngineURL(aType, aMethod, aTemplate) {
  ENSURE_ARG(aType && aMethod && aTemplate,
             "missing type, method or template for EngineURL!");

  var method = aMethod.toUpperCase();
  var type   = aType.toLowerCase();

  ENSURE(method == "GET" || method == "POST",
         "method passed to EngineURL must be \"GET\" or \"POST\"");

  ENSURE(type == "text/html", "EngineURLs must be of type text/html!",
         Cr.NS_ERROR_NOT_IMPLEMENTED);

  ENSURE(!kIllegalWords.test(aTemplate), "Invalid URL parameter!",
         Cr.NS_ERROR_NOT_IMPLEMENTED);

  this.type     = type;
  this.method   = method;
  this.template = aTemplate;
  this.params   = [];
}
EngineURL.prototype = {
  addParam: function (aName, aValue) {
    this.params.push(new QueryParameter(aName, aValue));
  },
};


function QueryParameter(aName, aValue) {
  ENSURE_ARG(aName && aValue, "missing name or value for QueryParameter!");

  ENSURE(!kIllegalWords.test(aValue),
         "Illegal value while creating a QueryParameter",
         Cr.NS_ERROR_NOT_IMPLEMENTED)

  this.name = aName;
  this.value = aValue;
}

function ENSURE(assertion, message, resultCode) {
  if (!assertion) {
    LOG(message, true);
    throw resultCode;
  }
}
function ENSURE_ARG(assertion, message) {
  ENSURE(assertion, message, Cr.NS_ERROR_INVALID_ARG);
}

const DEFAULT_QUERY_CHARSET = "ISO-8859-1";

const kCharsetCodes = [];
kCharsetCodes[0] = "x-mac-roman";
kCharsetCodes[6] = "x-mac-greek";
kCharsetCodes[35] = "x-mac-turkish";
kCharsetCodes[513] = "ISO-8859-1";
kCharsetCodes[514] = "ISO-8859-2";
kCharsetCodes[517] = "ISO-8859-5";
kCharsetCodes[518] = "ISO-8859-6";
kCharsetCodes[519] = "ISO-8859-7";
kCharsetCodes[520] = "ISO-8859-8";
kCharsetCodes[521] = "ISO-8859-9";
kCharsetCodes[1049] = "IBM864";
kCharsetCodes[1280] = "windows-1252";
kCharsetCodes[1281] = "windows-1250";
kCharsetCodes[1282] = "windows-1251";
kCharsetCodes[1283] = "windows-1253";
kCharsetCodes[1284] = "windows-1254";
kCharsetCodes[1285] = "windows-1255";
kCharsetCodes[1286] = "windows-1256";
kCharsetCodes[1536] = "us-ascii";
kCharsetCodes[1584] = "GB2312";
kCharsetCodes[1585] = "x-gbk";
kCharsetCodes[1600] = "EUC-KR";
kCharsetCodes[2080] = "ISO-2022-JP";
kCharsetCodes[2096] = "ISO-2022-CN";
kCharsetCodes[2112] = "ISO-2022-KR";
kCharsetCodes[2336] = "EUC-JP";
kCharsetCodes[2352] = "GB2312";
kCharsetCodes[2353] = "x-euc-tw";
kCharsetCodes[2368] = "EUC-KR";
kCharsetCodes[2561] = "Shift_JIS";
kCharsetCodes[2562] = "KOI8-R";
kCharsetCodes[2563] = "Big5";
kCharsetCodes[2565] = "HZ-GB-2312";

function getCharSetFromCode(aCode) {
  if (kCharsetCodes[aCode])
    return kCharsetCodes[aCode];

  return DEFAULT_QUERY_CHARSET;
}

function makeURI(aURLSpec) {
  var ios = Cc["@mozilla.org/network/io-service;1"].
            getService(Ci.nsIIOService);
  try {
    return ios.newURI(aURLSpec, null, null);
  } catch (ex) { }

  return null;
}
